#include "driver.h"
#include "wpc.h"
#include "wpcsim.h"

/* 161000  Corrected a timer error. Was sometimes counting VBLANKS instead */
/*         of 1/10s */
/* 231000  Now uses textOutf function */
#define WPCSIM_SHOOTRELTIME   3 /* how long the fake solenoid is active VBLANKS */
#define WPCSIM_TIMEFACTOR     6 /* multiplier for VBLANK to simulator timing */
#define WPCSIM_MAXSPIN        2 /* spinners in parallell */

static void wpcsim_startSpin(int swNo, int time);
static void wpcsim_updateSpin(void);

/*---------------------
/  Global variables
/---------------------*/
static wpcsim_tSimData *simData = NULL;  /* data about the game simulator */
static struct {
  wpcsim_tBallStatus balls[WPCSIM_MAXBALLS]; /* ball status etc */
  int  currBall;      /* currently select ball */
  int  shooterRel;    /* plunger released */
  int  shooterSpeed;  /* */
  struct {
    int swNo;         /* spinner switch */
    int spinCount;    /* time until next spin */
    int nextSpin;     /* timer for next spin */
  } spinner[WPCSIM_MAXSPIN];
} locals;

/*-----------------------------------------
/  Run the simulator (called every VBLANK)
/-----------------------------------------*/
void wpcsim_run(int *inports, int useSimKeys, int noOfBalls) {
  int ii;

  for (ii = 0; ii < simData->inports; ii++)
    inports[WPC_CUSTINPORT+ii] = readinputport(WPC_CUSTINPORT+ii);

  /*----------------------
  /  Change active ball ?
  /-----------------------*/
  locals.balls[locals.currBall].current = FALSE;
    if (inports[WPC_SIMINPORT] & WPCSIM_NEXTKEY) locals.currBall += 1;
    if (inports[WPC_SIMINPORT] & WPCSIM_PREVKEY) locals.currBall -= 1;
    if (locals.currBall < 0)                     locals.currBall = noOfBalls -1;
    if (locals.currBall >= noOfBalls)            locals.currBall = 0;
  locals.balls[locals.currBall].current = TRUE;

  /* update spinners running */
  wpcsim_updateSpin();

  /*-------------------
  /  Check user input
  /--------------------*/
  if (useSimKeys) {
    wpcsim_tBallStatus *ball = &locals.balls[locals.currBall];
    wpcsim_tInportData *iData = simData->inportData;

    while (iData->mask) {
      if ((inports[iData->port + WPC_CUSTINPORT] & iData->mask) == iData->mask) {
        /*-- mask this event off --*/
        inports[iData->port + WPC_CUSTINPORT] ^= iData->mask;
        /*-- check all conditions --*/
        if ((iData->condition == WPCSIM_STANY) || (inports[WPC_SIMINPORT] & WPCSIM_IGNOREKEY) ||
            ((iData->condition == 0) && (ball->state == stFree)) ||
            ((iData->condition > 0) && (iData->condition < WPCSIM_STATES) &&
             (ball->state == stFree) && wpc_getSw(iData->condition)) ||
            ((iData->condition < 0) && simData->keyCond &&
             simData->keyCond(-iData->condition, ball->state, &inports[WPC_CUSTINPORT])) ||
            ((iData->condition >= WPCSIM_STATES) && (iData->condition == ball->state))) {
          /*-- condition met --*/
          if (iData->action >= WPCSIM_STATES) {
            /*-- change state --*/
            if ((ball->state == ball->nextState) || (ball->timer <= 0)) {
              ball->nextState = iData->action;
              ball->timer = ball->nextTimer = 1;
            }
          }
          else if (iData->action > 0) { /* activate switch */
            wpc_setSw(iData->action, TRUE);
          }
        }
      }
      else if ((iData->action > 0) && (iData->action < WPCSIM_STATES)) {
        wpc_setSw(iData->action, FALSE); /* reset switch regardless of ball state */
      }
      iData += 1;
    }
  }
  /*----------------------------------------------
  /  update swithes depending playfield mechanics
  /-----------------------------------------------*/
  if (simData->manShooter) {
    /*--  ball shooter --*/
    if (inports[WPC_SIMINPORT] & WPCSIM_SHOOTERKEY) {
      locals.shooterSpeed += 1;
      if (locals.shooterSpeed > 50)
        locals.shooterSpeed = 50;
    }
    else if (locals.shooterSpeed > 0) {
      /*-- Shooters been released --*/
      if (locals.shooterRel == 1) locals.shooterSpeed = 0;
      if (locals.shooterRel > 0)  locals.shooterRel -= 1;
      else                        locals.shooterRel = WPCSIM_SHOOTRELTIME;
    }
  }
  /*----------------------------------
  / Game specific playfield mechanics
  /-----------------------------------*/
  if (simData->handleMech) simData->handleMech(&inports[WPC_CUSTINPORT]);

  /*------------------------
  /  update ball positions
  /-------------------------*/
  if (simData->stateData) {
    for (ii = 0; ii < WPCSIM_MAXBALLS; ii++) {
      wpcsim_tBallStatus *ball = &locals.balls[ii];
      if (ii >= noOfBalls) {
        /* this ball is not in the game */
        ball->nextState = stNotInst;
        ball->timer = 0;
      }
      if (ball->timer > 0)
        ball->timer -= 1;
      else if (ball->timer < 0) {
        ball->timer = ball->nextTimer;
        ball->state = stPending;
      }
      while (ball->timer == 0) {
        wpcsim_tState *sd = &simData->stateData[WPCSIM_STATENO(ball->nextState)];
        if (ball->state != ball->nextState) {
          /*-- try to move the ball --*/
          if ((sd->solNo) && wpc_getSol(sd->solNo)) {
            /*-- next state is blocked by a solenoid --*/
            break;
          }
          if (!(sd->flags & WPCSIM_STNOTEXCL)) {
            int jj;
            /*-- next state is exclusive --*/
            /*-- check if any other ball is in that state --*/
            for (jj = 0; jj < WPCSIM_MAXBALLS; jj++) {
              if ((jj != ii) &&
                  ((locals.balls[jj].state == ball->nextState) ||
                   ((ball->state != stPending) &&
                    (locals.balls[jj].nextState == ball->nextState) &&
                    (locals.balls[jj].state == stPending)))) {
                /*-- next state is occupied --*/
                goto nextStateBusy;
              }
            }
          }
          /*-- next state OK --*/
          if (ball->state == stPending) {
            /*-- go into the new state --*/
            ball->timer = sd->timer * WPCSIM_TIMEFACTOR;
            if (sd->swNo) { /* if the new state includes a switch, activate it */
              if (sd->flags & WPCSIM_STSPINNER)
                wpcsim_startSpin(sd->swNo, WPCSIM_SPINSEC(inports[WPC_SIMINPORT]));
              else
                wpc_setSw(sd->swNo, !((sd->flags & WPCSIM_STNOTBALL(0)) && (sd->flags & ball->type)));
            }
            ball->state = ball->nextState;
          }
          else {
            /*-- leave the current state first --*/
            ball->timer = -1;
            /*-- if the old state includes a switch, deactivate it --*/
            sd = &simData->stateData[WPCSIM_STATENO(ball->state)];
            if (!(sd->flags & (WPCSIM_STSWKEEP | WPCSIM_STSPINNER)) && sd->swNo)
              wpc_setSw(sd->swNo, FALSE);
            ball->state = stPending;
          }
        }
        else if ((sd->solNo == 0) || (sd->flags & WPCSIM_STIGNORESOL) || wpc_getSol(sd->solNo)) {
          if (sd->flags & WPCSIM_STSHOOT) ball->speed = locals.shooterSpeed;
          if (sd->nextState) {
            ball->nextState = ((sd->altSol) && wpc_getSol(sd->altSol)) ?
                              sd->altState : sd->nextState;
            ball->nextTimer = sd->nextTimer * WPCSIM_TIMEFACTOR;
          }
          else if ((simData->handleBallState == NULL) ||
                   !simData->handleBallState(ball, &inports[WPC_CUSTINPORT])) {
            /*-- The ball stays in its current state --*/
            break;
          }
        }
        else
          break;
      }
nextStateBusy: ;
    }
  }
}

/*-----------------------
/  Draw simulator data
/  (ball status etc.)
/------------------------*/
void wpcsim_draw(void) {
  struct rectangle rec;
  int ii;

  /*-- draw ball status --*/
  if (simData->stateData) {
    for (ii = 0; ii < WPCSIM_MAXBALLS; ii++)
      if (locals.balls[ii].state != stPending)
        wpc_textOutf(160, 10*ii+10, locals.balls[ii].current ? WHITE : BLACK,
                     "%c %-20s", locals.balls[ii].type ? '*' : ' ',
                      simData->stateData[WPCSIM_STATENO(locals.balls[ii].state)].name);
  }
  if (simData->manShooter) {
    /*-- draw shooter position --*/
    rec.min_y = 250; rec.max_y = 255;
    rec.min_x = 10; rec.max_x = 10+locals.shooterSpeed/2;
    fillbitmap(Machine->scrbitmap, Machine->pens[WHITE], &rec);
    rec.min_x = 11+locals.shooterSpeed/2; rec.max_x = 35;
    fillbitmap(Machine->scrbitmap, Machine->pens[BLACK], &rec);
  }
}

/*--------------------
/ Simulated solenoids
/---------------------*/
int wpcsim_getSol(int solNo) {
  if (solNo == sShooterRel)
    return locals.shooterRel > 0;
  return 0;
}

/*--------------------------
/  initialize the simluator
/---------------------------*/
int wpcsim_init(wpcsim_tSimData *gameSimData, int *inports) {
  int ii;

  simData = gameSimData;

  memset(locals.balls, 0, sizeof(locals.balls));

  for (ii = 0; ii < simData->inports; ii++)
    inports[WPC_CUSTINPORT+ii] = readinputport(WPC_CUSTINPORT+ii);

  for (ii = 0; ii < WPCSIM_BALLS(inports[WPC_SIMINPORT]); ii++) {
    locals.balls[ii].nextState = simData->ballStart[ii];
    locals.balls[ii].state = stPending;
  }
  if (simData->initSim)
    simData->initSim(locals.balls, &inports[WPC_CUSTINPORT], WPCSIM_BALLS(inports[WPC_SIMINPORT]));
  return (simData->inportData != NULL);
}

/*------------------
/  Spinner handling
/-------------------*/
static void wpcsim_startSpin(int swNo, int time) {
  static int spinStart[] = {0,251,219,191,172,143,114,87,65};
  int idx = -1;
  int ii;

  for (ii = 0; ii < WPCSIM_MAXSPIN; ii++)
    if (locals.spinner[ii].swNo == swNo)
      { idx = ii; break; } /* already running */
    else if (locals.spinner[ii].swNo == 0)
      idx = ii;
  if (idx < 0)
    { logerror("Only %d spinners supported\n",WPCSIM_MAXSPIN); return; }

  locals.spinner[idx].swNo = swNo;
  locals.spinner[idx].spinCount = 0;
  locals.spinner[idx].nextSpin = spinStart[time];
  wpc_setSw(swNo, TRUE);
}

static void wpcsim_updateSpin(void) {
  int ii;

  for (ii = 0; ii < WPCSIM_MAXSPIN; ii++)
    if (locals.spinner[ii].swNo) {
      if (locals.spinner[ii].spinCount == locals.spinner[ii].nextSpin) {
        /* activate spinner switch and start new timer */
        locals.spinner[ii].spinCount = 0;
        locals.spinner[ii].nextSpin = locals.spinner[ii].nextSpin * 20 / 19;
        if (locals.spinner[ii].nextSpin <= 275 )
          wpc_setSw(locals.spinner[ii].swNo, TRUE);
        else /* spinner has stopped */
          locals.spinner[ii].swNo = 0;
      }
      else if (locals.spinner[ii].spinCount == 30) /* activated time */
        wpc_setSw(locals.spinner[ii].swNo, FALSE);
      locals.spinner[ii].spinCount += 10;
    }
}

