#include <stdarg.h>
#include "driver.h"
#include "cpu/m6809/m6809.h"
#include "wpcsound.h"
#include "dcs.h"
#include "wpcsim.h"
#include "wpc.h"
#include "sound/mixer.h"
#include "snd_cmd.h"

/* 201000  Added "antialiasing" to the DMD. */
/* 221000  Added wpc_textOutf */
/* 281000  Added clearing of variables at startup (needed for WIN32 and MAC) */

#define ANTIALIAS
#define WPC_COLOR(x)  Machine->pens[(x)]

#define WPC_VBLANKFREQ    60 /* VBLANK frequency */

#define WPC_VBLANKDIV      4 /* How often to check the DMD FIRQ interrupt */
#define WPC_SOLSMOOTH      4 /* Smooth the Solenoids over this numer of VBLANKS */
#define WPC_LAMPSMOOTH     2 /* Smooth the lamps over this number of VBLANKS */
#define WPC_FLIPSTROKETIME 2 /* Timer for flipper to reach top VBLANKs */

/*-- IRQ frequence, most WPC functions are performed at 1/16 of this frequency --*/
#define WPC_IRQFREQ      1000 /* IRQ Frequency */

/*-- no of DMD frames to add together to create shades (hardcoded, do not change) --*/
#define DMD_FRAMES         3

#define WPC_FIRSTSIMROW   80 /* first free row on display */

/*-- default screen size */
#define WPC_SCREENX 256
#define WPC_SCREENY 256

/*---------------------
/  local WPC functions
/----------------------*/
/*-- Display handling --*/
static void wpc_initpalette(unsigned char *game_palette, unsigned short *game_colortable,
                            const unsigned char *color_prom);
static int  dmd_start(void);
static void dmd_refresh(struct osd_bitmap *bitmap,int full_refresh);
static void alpha_refresh(struct osd_bitmap *bitmap,int full_refresh);
static void drawStatus(struct osd_bitmap *bitmap);

/*-- interrupt handling --*/
static int wpc_vblank(void);
static int wpc_irq(void);

/*-- PIC security chip emulation --*/
static int wpc_pic_r(void);
static void wpc_pic_w(int data);
static void wpc_serialCnv(char no[21], unsigned char schip[16],
                          unsigned char code[3]);

/*-- misc. --*/
static void wpc_init(void);
static void wpc_updateSw(void);
static void wpc_nvram(void *file, int write);

/*------------------------
/  DMD display registers
/-------------------------*/
#define DMD_PAGE3A00    (0x3fbc - WPC_BASE)
#define DMD_PAGE3800    (0x3fbe - WPC_BASE)
#define DMD_VISIBLEPAGE (0x3fbf - WPC_BASE)
#define DMD_FIRQLINE    (0x3fbd - WPC_BASE)

/*---------------------
/  Global variables
/---------------------*/
wpc_tGameData *wpc_gameData = NULL;  /* data about the running game */
unsigned char *wpc_data;	     /* WPC registers */
wpc_tGlobals wpc_globals;
int WPC_gWPC95;      /* dcs95 sound, used in ADSP2100 patch ? */

/*--------
/ Palette
/---------*/
static unsigned char wpc_palette[] = {
/*  0 */ 0x00,0x00,0x00, /* Background */
/*  1 */ 0x30,0x00,0x00, /* "Black" Dot    */
/*  2 */ 0x5f,0x4a,0x0a, /* Intensity  33% */
/*  3 */ 0xaf,0x95,0x15, /* Intensity  66% */
/*  4 */ 0xff,0xe0,0x20, /* Intensity 100% */
/*  5 */ 0x40,0x40,0x40, /* Grey */
/*  6 */ 0x10,0x40,0x00, /* Dark Green */
/*  7 */ 0xff,0xe0,0x00, /* Intensity 100% (again) */
/*  8 */ 0x40,0x00,0x00, /* dark Red */
/*  9 */ 0x40,0x20,0x00, /* dark orange */
/* 10 */ 0x40,0x40,0x00, /* dark yellow */
/* 11 */ 0xff,0xff,0xff, /* White */
/* 12 */ 0x40,0xff,0x00, /* green */
/* 13 */ 0x00,0x00,0x00, /* not used */
/* 14 */ 0xff,0x00,0x00, /* Red */
/* 15 */ 0xff,0x80,0x00, /* orange */
/* 16 */ 0xff,0xff,0x00  /* yellow */
#ifdef ANTIALIAS
/* 17 */,0x00,0x00,0x00,
/* 18 */ 0x00,0x00,0x00,
/* 19 */ 0x00,0x00,0x00,
/* 20 */ 0x00,0x00,0x00,
/* 21 */ 0x00,0x00,0x00,
/* 22 */ 0x00,0x00,0x00,
/* 23 */ 0x00,0x00,0x00
#endif /* ANTIALIAS */
};

#define DMD_DOTOFF  1
#define DMD_DOT33   2
#define DMD_DOT66   3
#define DMD_DOTON   7

/*------------------
/  local variables
/-------------------*/
static struct {
  int  solData;        /* current value of solenoids 1-28 */
  int  solFlip;        /* current value of flipper solenoids */
  int  flipTimer[4];   /* time since flipper was activated (used for EOS simulation) */
  unsigned char invSw[WPC_MAXSWCOL];  /* Active low switches */

  char lampMatrix[WPC_MAXLAMPCOL];    /* Curren lamp matrix state */
  int  vblankCount;                   /* vblank interrupt counter */
  struct {
    unsigned char sData[16];
    unsigned char codeNo[3];
    unsigned char lastW;             /* last written command */
    unsigned char sNoS;              /* serial number scrambler */
    unsigned char count;
    int           codeW;
  } pic;
  int simAvail;            /* simulator (keys) available */
  int DMDsize;             /* 1=compact, 2=Normal */
  int time, ltime, ticks;
  int pageMask;            /* page handling */
  struct { unsigned char hi, lo; } alphaSeg[2][16];
  int alphaLoad;           /* counter for loading of alpha segments */
  int dmdOnly;             /* draw only the dmd (or alpha segemnts) */
} locals;

/*-- pointers --*/
static unsigned char *DMDFrames[DMD_FRAMES];
static int nextDMDFrame;
static void *wpc_printfile = NULL;

/*---------------------------
/  Memory map for CPU board
/----------------------------*/
static MEMORY_READ_START(wpc_readmem)
  { 0x0000, 0x37ff, MRA_RAM },
  { 0x3800, 0x39ff, MRA_BANK2 },  /* DMD */
  { 0x3A00, 0x3bff, MRA_BANK3 },  /* DMD */
  { 0x3c00, 0x3faf, MRA_RAM },
  { 0x3fb0, 0x3fff, wpc_r },
  { 0x4000, 0x7fff, MRA_BANK1 },
  { 0x8000, 0xffff, MRA_ROM },
MEMORY_END

static MEMORY_WRITE_START(wpc_writemem)
  { 0x0000, 0x37ff, MWA_RAM },
  { 0x3800, 0x39ff, MWA_BANK2 },
  { 0x3A00, 0x3bff, MWA_BANK3 },
  { 0x3c00, 0x3faf, MWA_RAM },
  { 0x3fb0, 0x3fff, wpc_w, &wpc_data },
MEMORY_END

/*-----------------
/  Machine drivers
/------------------*/
struct MachineDriver machine_driver_wpcDMD = {
  {
    {
      CPU_M6809, 2000000, /* 2 Mhz? */
      wpc_readmem, wpc_writemem, NULL, NULL,
      wpc_vblank, WPC_VBLANKDIV,
      wpc_irq, WPC_IRQFREQ
    }
  },
  WPC_VBLANKFREQ, DEFAULT_60HZ_VBLANK_DURATION,
  100,
  wpc_init,
  WPC_SCREENX, WPC_SCREENY, { 0, WPC_SCREENX-1, 0, WPC_SCREENY-1 },
  0, sizeof(wpc_palette)/sizeof(wpc_palette[0])/3, 0, wpc_initpalette,
  VIDEO_SUPPORTS_DIRTY | VIDEO_TYPE_RASTER,
  0,
  dmd_start, NULL, dmd_refresh,
  0,0,0,0, {{ 0,0 }},
  wpc_nvram
};

struct MachineDriver machine_driver_wpcAlpha = {
  {
    {
      CPU_M6809, 2000000, /* 2 Mhz? */
      wpc_readmem, wpc_writemem, NULL, NULL,
      wpc_vblank, WPC_VBLANKDIV,
      wpc_irq, WPC_IRQFREQ
    }
  },
  WPC_VBLANKFREQ, DEFAULT_60HZ_VBLANK_DURATION,
  100,
  wpc_init,
  WPC_SCREENX, WPC_SCREENY, { 0, WPC_SCREENX-1, 0, WPC_SCREENY-1 },
  0, sizeof(wpc_palette)/sizeof(wpc_palette[0])/3, 0, wpc_initpalette,
  VIDEO_SUPPORTS_DIRTY | VIDEO_TYPE_RASTER,
  0,
  NULL, NULL, alpha_refresh,
  0,0,0,0, {{ 0,0 }},
  wpc_nvram
};

#ifdef WPC_WPCSOUND
struct MachineDriver machine_driver_wpcDMD_s = {
  {
    {
      CPU_M6809, 2000000, /* 2 Mhz? */
      wpc_readmem, wpc_writemem, NULL, NULL,
      wpc_vblank, WPC_VBLANKDIV,
      wpc_irq, WPC_IRQFREQ
    }
    WPCS_SOUNDCPU
  },
  WPC_VBLANKFREQ, DEFAULT_60HZ_VBLANK_DURATION,
  100,
  wpc_init,
  WPC_SCREENX, WPC_SCREENY, { 0, WPC_SCREENX-1, 0, WPC_SCREENY-1 },
  0, sizeof(wpc_palette)/sizeof(wpc_palette[0])/3, 0, wpc_initpalette,
  VIDEO_SUPPORTS_DIRTY | VIDEO_TYPE_RASTER,
  0,
  dmd_start, NULL, dmd_refresh,
  0, 0, 0, 0, { WPCS_SOUND },
  wpc_nvram
};
struct MachineDriver machine_driver_wpcAlpha_s = {
  {
    {
      CPU_M6809, 2000000, /* 2 Mhz? */
      wpc_readmem, wpc_writemem, NULL, NULL,
      wpc_vblank, WPC_VBLANKDIV,
      wpc_irq, WPC_IRQFREQ
    }
    WPCS_SOUNDCPU
  },
  WPC_VBLANKFREQ, DEFAULT_60HZ_VBLANK_DURATION,
  100,
  wpc_init,
  WPC_SCREENX, WPC_SCREENY, { 0, WPC_SCREENX-1, 0, WPC_SCREENY-1 },
  0, sizeof(wpc_palette)/sizeof(wpc_palette[0])/3, 0, wpc_initpalette,
  VIDEO_SUPPORTS_DIRTY | VIDEO_TYPE_RASTER,
  0,
  NULL, NULL, alpha_refresh,
  0, 0, 0, 0, { WPCS_SOUND },
  wpc_nvram
};
#endif /* WPC_WPCSOUND */

#ifdef WPC_DCSSOUND
struct MachineDriver machine_driver_wpcDCS_s = {
  {
    {
      CPU_M6809, 2000000, /* 2 Mhz? */
      wpc_readmem, wpc_writemem, NULL, NULL,
      wpc_vblank, WPC_VBLANKDIV,
      wpc_irq, WPC_IRQFREQ
    }
    DCS1_SOUNDCPU
  },
  WPC_VBLANKFREQ, DEFAULT_60HZ_VBLANK_DURATION,
  500,
  wpc_init,
  WPC_SCREENX, WPC_SCREENY, { 0, WPC_SCREENX-1, 0, WPC_SCREENY-1 },
  0, sizeof(wpc_palette)/sizeof(wpc_palette[0])/3, 0, wpc_initpalette,
  VIDEO_SUPPORTS_DIRTY | VIDEO_TYPE_RASTER,
  0,
  dmd_start, NULL, dmd_refresh,
  0, 0, 0, 0, { DCS_SOUND },
  wpc_nvram
};

struct MachineDriver machine_driver_wpcDCS95_s = {
  {
    {
      CPU_M6809, 2000000, /* 2 Mhz? */
      wpc_readmem, wpc_writemem, NULL, NULL,
      wpc_vblank, WPC_VBLANKDIV,
      wpc_irq, WPC_IRQFREQ
    }
    DCS2_SOUNDCPU
  },
  WPC_VBLANKFREQ, DEFAULT_60HZ_VBLANK_DURATION,
  500,
  wpc_init,
  WPC_SCREENX, WPC_SCREENY, { 0, WPC_SCREENX-1, 0, WPC_SCREENY-1 },
  0, sizeof(wpc_palette)/sizeof(wpc_palette[0])/3, 0, wpc_initpalette,
  VIDEO_SUPPORTS_DIRTY | VIDEO_TYPE_RASTER,
  0,
  dmd_start, NULL, dmd_refresh,
  0, 0, 0, 0, { DCS_SOUND },
  wpc_nvram
};
#endif /* WPC_DCSSOUND */

/*-------------------
/ display handling
/--------------------*/
/* alphanumeric display characters */
static UINT32 alphachar[20][17] = {
/*                       all        0001       0002       0004       0008       0010       0020       0040?      0080       0100       0200       0400       0800       1000       2000       4000       8000 */
/*  xxxxxxxxxxx    */{0x05555540,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x05555540,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000},
/* xx    x    xx   */{0x14010050,0x04000000,0x00010000,0x00000040,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x00000000,0x00000000,0x10000000,0x00000000,0x00000000},
/* x x   x   x x   */{0x11010110,0x01000000,0x00010000,0x00000100,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x00000000,0x00000000,0x10000000,0x00000000,0x00000000},
/* x x   x   x x   */{0x11010110,0x01000000,0x00010000,0x00000100,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x00000000,0x00000000,0x10000000,0x00000000,0x00000000},
/* x  x  x  x  x   */{0x10410410,0x00400000,0x00010000,0x00000400,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x00000000,0x00000000,0x10000000,0x00000000,0x00000000},
/* x  x  x  x  x   */{0x10410410,0x00400000,0x00010000,0x00000400,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x00000000,0x00000000,0x10000000,0x00000000,0x00000000},
/* x   x x x   x   */{0x10111010,0x00100000,0x00010000,0x00001000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x00000000,0x00000000,0x10000000,0x00000000,0x00000000},
/* x   x x x   x   */{0x10111010,0x00100000,0x00010000,0x00001000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x00000000,0x00000000,0x10000000,0x00000000,0x00000000},
/* x    xxx    x   */{0x10054010,0x00040000,0x00010000,0x00004000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x00000000,0x00000000,0x10000000,0x00000000,0x00000000},
/*  xxxxx xxxxx    */{0x05545540,0x00000000,0x00000000,0x00000000,0x00005540,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x05540000,0x00000000},
/* x    xxx    x   */{0x10054010,0x00000000,0x00000000,0x00000000,0x00000000,0x00004000,0x00010000,0x00040000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x10000000,0x00000000,0x00000000,0x00000000},
/* x   x x x   x   */{0x10111010,0x00000000,0x00000000,0x00000000,0x00000000,0x00001000,0x00010000,0x00100000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x10000000,0x00000000,0x00000000,0x00000000},
/* x   x x x   x   */{0x10111010,0x00000000,0x00000000,0x00000000,0x00000000,0x00001000,0x00010000,0x00100000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x10000000,0x00000000,0x00000000,0x00000000},
/* x  x  x  x  x   */{0x10410410,0x00000000,0x00000000,0x00000000,0x00000000,0x00000400,0x00010000,0x00400000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x10000000,0x00000000,0x00000000,0x00000000},
/* x  x  x  x  x   */{0x10410410,0x00000000,0x00000000,0x00000000,0x00000000,0x00000400,0x00010000,0x00400000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x10000000,0x00000000,0x00000000,0x00000000},
/* x x   x   x x   */{0x11010110,0x00000000,0x00000000,0x00000000,0x00000000,0x00000100,0x00010000,0x01000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x10000000,0x00000000,0x00000000,0x00000000},
/* x x   x   x x   */{0x11010110,0x00000000,0x00000000,0x00000000,0x00000000,0x00000100,0x00010000,0x01000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x10000000,0x00000000,0x00000000,0x00000000},
/* xx    x    xx x */{0x14010051,0x00000000,0x00000000,0x00000000,0x00000000,0x00000040,0x00010000,0x04000000,0x00000000,0x00000000,0x00000000,0x00000010,0x00000000,0x10000000,0x00000000,0x00000000,0x00000001},
/*  xxxxxxxxxxx  x */{0x05555541,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x05555540,0x00000000,0x00000000,0x00000000,0x00000001},
/*              x  */{0x00000004,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000000,0x00000004}
};

static void wpc_initpalette(unsigned char *game_palette, unsigned short *game_colortable,
                            const unsigned char *color_prom) {
  memcpy(game_palette, wpc_palette, sizeof(wpc_palette));
#ifdef ANTIALIAS
  if (options.flicker) {
    int rStart = wpc_palette[1*3+0];
    int gStart = wpc_palette[1*3+1];
    int bStart = wpc_palette[1*3+2];
    int rStep = (wpc_palette[4*3+0] * options.flicker / 0xff - rStart) / 6;
    int gStep = (wpc_palette[4*3+1] * options.flicker / 0xff - gStart) / 6;
    int bStep = (wpc_palette[4*3+2] * options.flicker / 0xff - bStart) / 6;
    int ii;
    for (ii = 18; ii < 24; ii++) {
      rStart += rStep; gStart += gStep; bStart += bStep;
      game_palette[ii*3+0] = rStart;
      game_palette[ii*3+1] = gStart;
      game_palette[ii*3+2] = bStart;
    }
  }
#endif /* ANTIALIAS */
}

static int dmd_start(void) {
  unsigned char *RAM = memory_region(WPC_MEMREG_DMD);
  int ii;

  for (ii = 0; ii < DMD_FRAMES; ii++)
    DMDFrames[ii] = RAM;
  nextDMDFrame = 0;
  return 0;
}

/*----------------------------
/ Update alphanumeric display
/-----------------------------*/
static void alpha_refresh(struct osd_bitmap *bitmap, int fullRefresh) {
  int ii, jj, kk, ll;

  /*Process manual sound commands*/
  if(USESOUND)
	  manual_sound_commands(bitmap);

  /* Drawing is not optimised so just clear everything */
  if (fullRefresh) fillbitmap(bitmap,Machine->pens[0],NULL);

  for (ii = 0; ii < 2; ii++) { /* 2 Rows */
    for (jj = 0; jj < 16; jj++) { /* 16 characters */
      int bits = (locals.alphaSeg[ii][jj].hi << 8) | (locals.alphaSeg[ii][jj].lo);
      UINT32 pixel[20];

      for (kk = 0; kk < 20; kk++)
        pixel[kk] = alphachar[kk][0];
      for (kk = 1; bits; kk++, bits >>= 1) { /* max 15 segments */
        if (bits & 0x01)
          for (ll = 0; ll < 20; ll++)
            pixel[ll] += alphachar[ll][kk];
      }
      for (kk = 0; kk < 20; kk++) {
        UINT8 *line = &bitmap->line[ii*40+kk][16*jj];
        UINT32 p = pixel[kk];

        for (ll = 0; ll < 15; ll++) { /* 15 pixels per row */
          *line++ = WPC_COLOR(((p >> 28) & 0x03)*3 & 0x05); /* 0->0, 1->1, 2->4 */
          p <<= 2;
        }
      }
    }
  }
  osd_mark_dirty(0,0,256,50);
  /*Don't draw status if in sound mode*/
  if(sound_mode==0)
	  if (!locals.dmdOnly) drawStatus(bitmap);
}

/*----------------------------------------------
/ Update DMD
/
/ Intensity depends on how many times the pixel
/ been on in the last 3 frames
/
/ This can probably be optimised
/-----------------------------------------------------*/
static void dmd_refresh(struct osd_bitmap *bitmap, int fullRefresh) {
  UINT8 **lines = bitmap->line;
#ifdef ANTIALIAS
  UINT8 dotCol[34][129];
#else /* ANTIALIAS */
  int step = locals.DMDsize;
#endif /* ANTIALIAS */
  int ii,jj,kk;

  /*Process manual sound commands*/
  if(USESOUND)
  	manual_sound_commands(bitmap);

  /* Drawing is not optimised so just clear everything */
  if (fullRefresh) fillbitmap(bitmap,Machine->pens[0],NULL);

  kk = 0;

#ifdef ANTIALIAS
  for (ii = 1; ii < 33; ii++) {
    UINT8 *line = &dotCol[ii][0];
    for (jj = 0; jj < 16; jj++) {
      unsigned int intens1 = ((DMDFrames[0][kk] & 0x55) + (DMDFrames[1][kk] & 0x55) +
                              (DMDFrames[2][kk] & 0x55));
      unsigned int intens2 = ((DMDFrames[0][kk] & 0xaa) + (DMDFrames[1][kk] & 0xaa) +
                              (DMDFrames[2][kk] & 0xaa));

      *line++ = (intens1)    & 0x03;
      *line++ = (intens2>>1) & 0x03;
      *line++ = (intens1>>2) & 0x03;
      *line++ = (intens2>>3) & 0x03;
      *line++ = (intens1>>4) & 0x03;
      *line++ = (intens2>>5) & 0x03;
      *line++ = (intens1>>6) & 0x03;
      *line++ = (intens2>>7) & 0x03;
      kk +=1;
    }
    *line = 0; /* to simplify antialiasing */
  }
  memset(&dotCol[33][0],0,sizeof(UINT8)*129);
  memset(&dotCol[0][0], 0,sizeof(UINT8)*129);

  for (ii = 0; ii < 33; ii++) {
    UINT8 *line = *lines++;
    if (ii > 0) {
      for (jj = 0; jj < 128; jj++) {
        *line++ = WPC_COLOR(dotCol[ii][jj] + DMD_DOTOFF);
        if (locals.DMDsize > 1)
          *line++ = WPC_COLOR(dotCol[ii][jj] + dotCol[ii][jj+1] + 17);
      }
    }
    if (locals.DMDsize > 1) {
      int col1 = dotCol[ii][0] + dotCol[ii+1][0];
      line = *lines++;
      for (jj = 0; jj < 128; jj++) {
        int col2 = dotCol[ii][jj+1] + dotCol[ii+1][jj+1];
        *line++ = WPC_COLOR(   col1           + 17);
        *line++ = WPC_COLOR(2*(col1 + col2)/5 + 17);
        col1 = col2;
      }
    }
  }
#else /* ANTIALIAS */

  for (ii = 0; ii < 32; ii++) {
    UINT8 *line = &lines[ii*locals.DMDsize][0];
    for (jj = 0; jj < 16; jj++) {
      unsigned int intens1 = ((DMDFrames[0][kk] & 0x55) + (DMDFrames[1][kk] & 0x55) +
                              (DMDFrames[2][kk] & 0x55));
      unsigned int intens2 = ((DMDFrames[0][kk] & 0xaa) + (DMDFrames[1][kk] & 0xaa) +
                              (DMDFrames[2][kk] & 0xaa));

      *line = WPC_COLOR(((intens1)    & 0x03) + DMD_DOTOFF); line += 2*step;
      *line = WPC_COLOR(((intens1>>2) & 0x03) + DMD_DOTOFF); line += 2*step;
      *line = WPC_COLOR(((intens1>>4) & 0x03) + DMD_DOTOFF); line += 2*step;
      *line = WPC_COLOR(((intens1>>6) & 0x03) + DMD_DOTOFF); line -= 5*step;
      *line = WPC_COLOR(((intens2>>1) & 0x03) + DMD_DOTOFF); line += 2*step;
      *line = WPC_COLOR(((intens2>>3) & 0x03) + DMD_DOTOFF); line += 2*step;
      *line = WPC_COLOR(((intens2>>5) & 0x03) + DMD_DOTOFF); line += 2*step;
      *line = WPC_COLOR(((intens2>>7) & 0x03) + DMD_DOTOFF); line += step;
      kk +=1;
    }
  }
#endif /* ANTIALIAS */
  osd_mark_dirty(0,0,128*locals.DMDsize,32*locals.DMDsize);
  
  /*Don't draw status if in sound mode*/
  if(sound_mode==0)
	  if (!locals.dmdOnly) drawStatus(bitmap);
}

static void drawStatus(struct osd_bitmap *bitmap) {
  UINT8 **lines = bitmap->line;
  int firstRow = WPC_FIRSTSIMROW;
  int ii, jj, bits;
  /*-----------------
  /  Draw the lamps
  /------------------*/
  if (wpc_gameData->lampData) {
    wpc_tDrawData *drawData = wpc_gameData->lampData;
    wpc_tDrawData *tmp = &drawData[2];
    UINT8 **line = &lines[WPC_FIRSTSIMROW+drawData[0].x];

    for (ii = 0; ii < WPC_CUSTLAMPCOL+wpc_gameData->lampCol; ii++) {
      bits = wpc_globals.lampMatrix[ii];
      for (jj = 0; jj < 8; jj++) {
        int color = (tmp->color == BLACK) ? DMD_DOTON : tmp->color;
        line[tmp->x][drawData[0].y + tmp->y] =
          WPC_COLOR((bits & 0x01) ? color : SHADE(color));
        bits >>= 1;
        tmp += 1;
      }
    }
    osd_mark_dirty(drawData[0].y,  drawData[0].x + WPC_FIRSTSIMROW,
                   drawData[0].y + drawData[1].y,
                   drawData[0].x + drawData[1].x + WPC_FIRSTSIMROW);

    firstRow = drawData[0].x + drawData[1].x + WPC_FIRSTSIMROW;
  }
  else {
    for (ii = 0; ii < WPC_CUSTLAMPCOL+wpc_gameData->lampCol; ii++) {
      UINT8 **line = &lines[firstRow];
      bits = wpc_globals.lampMatrix[ii];

      for (jj = 0; jj < 8; jj++) {
        line[0][ii*2] = WPC_COLOR((bits & 0x01) ? DMD_DOTON : SHADE(DMD_DOTON));
        line += 2;
        bits >>= 1;
      }
    }
    osd_mark_dirty(0,firstRow,16,firstRow+ii*2);
    firstRow += ii*2;
  } /* else */
  /*-------------------
  /  Draw the switches
  /--------------------*/
  firstRow += 5;
  if (wpc_gameData->swData) {
    wpc_tDrawData *drawData = wpc_gameData->swData;
    wpc_tDrawData *tmp = &drawData[2];
    UINT8 **line = &lines[WPC_FIRSTSIMROW+drawData[0].x];

    for (ii = 0; ii < WPC_CUSTSWCOL+wpc_gameData->swCol; ii++) {
      bits = wpc_globals.swMatrix[ii];
      for (jj = 0; jj < 8; jj++) {
        int color = (tmp->color == BLACK) ? DMD_DOTON : tmp->color;
        line[tmp->x][drawData[0].y + tmp->y] =
          WPC_COLOR((bits & 0x01) ? color : SHADE(color));
        tmp += 1;
        bits >>= 1;
      }
    }
    osd_mark_dirty(drawData[0].y,  drawData[0].x + WPC_FIRSTSIMROW,
                   drawData[0].y + drawData[1].y,
                   drawData[0].x + drawData[1].x + WPC_FIRSTSIMROW);
    if (drawData[0].x + drawData[1].x + WPC_FIRSTSIMROW > firstRow)
      firstRow = drawData[0].x + drawData[1].x + WPC_FIRSTSIMROW;
  }
  else {
    for (ii = 0; ii < WPC_CUSTSWCOL+wpc_gameData->swCol; ii++) {
      UINT8 **line = &lines[firstRow];
      bits = wpc_globals.swMatrix[ii];

      for (jj = 0; jj < 8; jj++) {
        line[0][ii*2] = WPC_COLOR((bits & 0x01) ? DMD_DOTON : SHADE(DMD_DOTON));
        line += 2;
        bits >>= 1;
      }
    }
    osd_mark_dirty(0,firstRow,16,firstRow+ii*2);
    firstRow += ii*2;
  } /* else */
  /*------------------------------
  /  Draw Solenoids and Flashers
  /-------------------------------*/
  firstRow += 5;
  if (wpc_gameData->solData) {
    wpc_tDrawData *drawData = wpc_gameData->solData;
    wpc_tDrawData *tmp = &drawData[2];
    UINT8 **line = &lines[WPC_FIRSTSIMROW+drawData[0].x];

    bits = locals.solData;
    for (ii = 0; ii < WPC_STDSOL; ii++) {
      int color = (tmp->color == BLACK) ? DMD_DOTON : tmp->color;
      line[tmp->x][drawData[0].y + tmp->y] =
        WPC_COLOR((bits & 0x01) ? color : SHADE(color));
      bits >>= 1;
      tmp += 1;
    }
    bits = locals.solFlip;
    for (ii = 0; ii < WPC_FLIPPERSOL; ii++) {
      int color = (tmp->color == BLACK) ? DMD_DOTON : tmp->color;
      line[tmp->x][drawData[0].y + tmp->y] =
        WPC_COLOR((bits & 0x01) ? color : SHADE(color));
      bits >>= 1;
      tmp += 1;
    }
    for (ii = 0; ii < wpc_gameData->custSol; ii++) {
      int color = (tmp->color == BLACK) ? DMD_DOTON : tmp->color;
      line[tmp->x][drawData[0].y + tmp->y] =
        WPC_COLOR((wpc_gameData->getSol(WPC_CUSTSOL+ii)) ? color : SHADE(color));
      tmp += 1;
    }
    osd_mark_dirty(drawData[0].y,  drawData[0].x + WPC_FIRSTSIMROW,
                   drawData[0].y + drawData[1].y,
                   drawData[0].x + drawData[1].x + WPC_FIRSTSIMROW);
  }
  else {
    UINT8 **line = &lines[firstRow];
    bits = locals.solData;
    for (ii = 0; ii < WPC_STDSOL; ii++) {
      line[0][(ii & 0x07) * 2] =
        WPC_COLOR((bits & 0x01) ? DMD_DOTON : SHADE(DMD_DOTON));
      bits >>= 1;
      line += ((ii & 0x07) == 0x07)*2;
    }
    line += 2;
    bits = locals.solFlip;
    for (ii = 0; ii < 8; ii++) {
      line[0][ii*2] =
        WPC_COLOR((bits & 0x01) ? DMD_DOTON : SHADE(DMD_DOTON));
      bits >>= 1;
    }
    line += 2;
    for (ii = 0; ii < wpc_gameData->custSol; ii++) {
      line[0][(ii & 0x07) * 2] =
        WPC_COLOR((wpc_gameData->getSol(WPC_CUSTSOL+ii)) ? DMD_DOTON : SHADE(DMD_DOTON));
      line += ((ii & 0x07) == 0x07)*2;
    }
    osd_mark_dirty(0,firstRow,16,firstRow+((WPC_STDSOL+7)/8+1+(wpc_gameData->custSol+7)/8)*2);
  } /* else */
  if (locals.simAvail) wpcsim_draw();
  /*-- draw game specific mechanics --*/
  if (wpc_gameData->drawMech)
    wpc_gameData->drawMech(&bitmap->line[WPC_FIRSTSIMROW]);
}

/*--------------------------------------------------------------
/ This is generated WPC_VBLANKDIV times per frame
/ = every 32/WPC_VBLANKDIV lines.
/ Generate a FIRQ if it matches the DMD line
/ Also do the smoothing of the solenoids and lamps
/--------------------------------------------------------------*/
static int wpc_vblank(void) {
  locals.vblankCount = (locals.vblankCount+1) % 16;

  if (wpc_gameData->wpcGen >= WPC_GENDMD) {
    /*-- check if the DMD line matches the requested interrupt line */
    if ((locals.vblankCount % WPC_VBLANKDIV) == (wpc_data[DMD_FIRQLINE]*WPC_VBLANKDIV/32)) {
      wpc_data[WPC_FIRQSRC] &= 0x7f; /* mark that this is a DMD interrupt */
      cpu_set_irq_line(WPC_CPUNO, M6809_FIRQ_LINE, HOLD_LINE);
    }
    if ((locals.vblankCount % WPC_VBLANKDIV) == 0) {
      /*-- This is the real VBLANK interrupt --*/
      DMDFrames[nextDMDFrame] = memory_region(WPC_MEMREG_DMD)+ (wpc_data[DMD_VISIBLEPAGE] & 0x0f) * 0x200;
      nextDMDFrame = (nextDMDFrame + 1) % DMD_FRAMES;
    }
  }

  /*------------------------------
  /  Update switches every vblank
  /-------------------------------*/
  if ((locals.vblankCount % WPC_VBLANKDIV) == 0) {
    /*-- update switches --*/
    wpc_updateSw();
  }
  /*--------------------------------------------------------
  /  Most solonoids don't have a holding coil so the software
  /  simulates it by pulsing the power to the main (and only) coil.
  /  (I assume this is why the Auto Fire diverter in TZ seems to flicker.)
  /  I simulate the coil position by looking over WPC_SOLSMOOTH vblanks
  /  and only turns the solenoid off if it has not been pulsed
  /  during that time.
  /-------------------------------------------------------------*/
  if ((locals.vblankCount % (WPC_VBLANKDIV*WPC_SOLSMOOTH)) == 0) {
    locals.solData = (wpc_data[WPC_SOLENOID1] << 24) |
                     (wpc_data[WPC_SOLENOID3] << 16) |
                     (wpc_data[WPC_SOLENOID4] <<  8) |
                      wpc_data[WPC_SOLENOID2];
    wpc_data[WPC_SOLENOID1] = wpc_data[WPC_SOLENOID2] = 0;
    wpc_data[WPC_SOLENOID3] = wpc_data[WPC_SOLENOID4] = 0;
  }
  /*--------------------------------------------------------
  / The lamp circuit pulses the power to the lamps one column
  / at a time. By only reading the value every WPC_LAMPSMOOTH
  / vblank we get a steady light.
  /
  / Williams changed the software for the lamphandling at some stage
  / earlier code (TZ 9.2)	newer code (TZ 9.4)
  / ------------		            -----------
  / Activate Column             Activate rows
  / Activate rows               Activate column
  / wait                        wait
  / Deactivate rows             repeat for next column
  / repeat for next column
  / ..
  / For the game it doesn't really matter but it confused me when
  / the lamp code here worked for 9.2 but not in 9.4.
  /-------------------------------------------------------------*/
  if ((locals.vblankCount % (WPC_VBLANKDIV*WPC_LAMPSMOOTH)) == 0) {
    memcpy(wpc_globals.lampMatrix, locals.lampMatrix, sizeof(wpc_globals.lampMatrix));
    memset(locals.lampMatrix, 0, sizeof(wpc_globals.lampMatrix));
  }
  return 0;
}


/*----------------------
/ Emulate the WPC chip
/-----------------------*/
READ_HANDLER(wpc_r) {
  int ii;

  switch (offset) {
    case WPC_FLIPPERS: /* Flipper switches */
      if (wpc_gameData->wpcGen < WPC_GEN95)
        return ~wpc_globals.swMatrix[WPC_CABINETSWCOL];
      break;
    case WPC_FLIPPERSW95:
      if (wpc_gameData->wpcGen >= WPC_GEN95)
        return ~wpc_globals.swMatrix[WPC_CABINETSWCOL];
      break;
    case WPC_SWCOINDOOR: /* cabinet switches */
      return wpc_globals.swMatrix[WPC_COINDOORSWCOL];
    case WPC_SWROWREAD: /* read row */
      if (wpc_gameData->wpcGen >= WPC_GENSECURITY)
        return wpc_pic_r();
      /*-- find the enabled column --*/
      for (ii = 0; ii < 8; ii++)
        if (((wpc_data[WPC_SWCOLSELECT] & (1<<ii))) > 0)
          return wpc_globals.swMatrix[ii+1];
      return 0; /* No column enabled */
    case WPC_SHIFTADRH:
        return wpc_data[WPC_SHIFTADRH] +
               ((wpc_data[WPC_SHIFTADRL] + (wpc_data[WPC_SHIFTBIT]>>3))>>8);
        break;
    case WPC_SHIFTADRL:
      return (wpc_data[WPC_SHIFTADRL] + (wpc_data[WPC_SHIFTBIT]>>3)) & 0xff;
    case WPC_SHIFTBIT:
    case WPC_SHIFTBIT2:
      return 1<<(wpc_data[offset] & 0x07);
    case WPC_FIRQSRC: /* FIRQ source */
      break;
    case WPC_DIPSWITCH:
      return WPC_DIPSW(readinputport(WPC_DIPINPORT));
    case WPC_RTCHOUR:
      return (locals.time/60) % 24;
    case WPC_RTCMIN:
      return locals.time % 60;
    case WPC_WATCHDOG:
      break;
    case WPC_SOUNDIF:
      if (USESOUND) {
        DBGLOG(("soundif_r: (PC=%4x)\n",cpu_get_pc()));
        return (wpc_gameData->wpcGen < WPC_GENDCS) ? wpcs_data_r(0)    : dcs_data_r(0);
      }
      return 0;
    case WPC_SOUNDBACK:
      if (USESOUND) {
        DBGLOG(("soundback_r: (PC=%4x)\n",cpu_get_pc()));
        return (wpc_gameData->wpcGen < WPC_GENDCS) ? wpcs_ctrl_r(0) : dcs_ctrl_r(0);
      }
      return 0;
    case WPC_PRINTBUSY:
      return 0;
    default:
      /*logerror("wpc_r %4x\n", offset+WPC_BASE);*/
      break;
  }
  return wpc_data[offset];
}

WRITE_HANDLER(wpc_w) {
  int tmp;
  int ii;

  switch (offset) {
    case WPC_ROMBANK: { /* change rom bank */
      int bank = data & locals.pageMask;
      cpu_setbank(1, memory_region(WPC_MEMREG_ROM)+ bank * 0x4000);
      break;
    }
    case WPC_FLIPPERS: /* Flipper coils */
      if (wpc_gameData->wpcGen < WPC_GEN95)
        locals.solFlip = ~data;
      break;
    case WPC_FLIPPERCOIL95:
      if (wpc_gameData->wpcGen >= WPC_GEN95)
        locals.solFlip = data;
      else if ((wpc_gameData->wpcGen == WPC_GENALPHA) && locals.alphaLoad) {
        locals.alphaSeg[1][wpc_data[WPC_ALPHAPOS]].hi = data;
        locals.alphaLoad -= 1;
      }
      break;
    /*-----------------------------
    /  Lamp handling
    /  Lamp values are accumulated
    /  see vblank for description
    /------------------------------*/
    case WPC_LAMPROW: /* row and column can be written in any order */
      tmp = wpc_data[WPC_LAMPCOLUMN];
      for (ii = 0; ii < 8; ii++) {
        if (tmp & 0x01)
          locals.lampMatrix[ii] |= data;
        tmp >>= 1;
      }
      break;
    case WPC_LAMPCOLUMN: /* row and column can be written in any order */
      tmp = data;
      for (ii = 0; ii < 8; ii++) {
        if (tmp & 0x01)
          locals.lampMatrix[ii] |= wpc_data[WPC_LAMPROW];
        tmp >>= 1;
      }
      break;
    case WPC_SWCOLSELECT:
      if (wpc_gameData->wpcGen >= WPC_GENSECURITY)
        wpc_pic_w(data);
      break;
    case WPC_GILAMPS:
      break; /* not simulated yet */
    case WPC_EXTBOARD1: /* WPC_ALPHAPOS */
      locals.alphaLoad = 4;
      break; /* just save position */
    case WPC_EXTBOARD2: /* WPC_ALPHA1 */
      if ((wpc_gameData->wpcGen == WPC_GENALPHA) && locals.alphaLoad) {
        locals.alphaSeg[0][wpc_data[WPC_ALPHAPOS]].hi = data;
        locals.alphaLoad -= 1;
      }
      break;
    case WPC_EXTBOARD3:
      if ((wpc_gameData->wpcGen == WPC_GENALPHA) && locals.alphaLoad) {
        locals.alphaSeg[0][wpc_data[WPC_ALPHAPOS]].lo = data;
        locals.alphaLoad -= 1;
      }
      break;
    /* case WPC_EXTBOARD4: */
    case WPC_EXTBOARD5:
      if ((wpc_gameData->wpcGen == WPC_GENALPHA) && locals.alphaLoad) {
        locals.alphaSeg[1][wpc_data[WPC_ALPHAPOS]].lo = data;
        locals.alphaLoad -= 1;
      }
      break;
    case WPC_SHIFTADRH:
    case WPC_SHIFTADRL:
    case WPC_SHIFTBIT:
    case WPC_DIPSWITCH:
      break; /* just save value */
    /*-------------------------------------------
    /  The solenoids are accumulated over time.
    /  See vblank interrupt for explanation.
    /--------------------------------------------*/
    case WPC_SOLENOID1:
    case WPC_SOLENOID2:
    case WPC_SOLENOID3:
    case WPC_SOLENOID4:
      data |= wpc_data[offset];
      break;
    case WPC_SOUNDIF:
      DBGLOG(("Sound cmd: %2x\n",data));
      if (USESOUND) {
        if (wpc_gameData->wpcGen < WPC_GENDCS) wpcs_data_w(0,data);
        else                                   dcs_data_w(0,data);
      }
      break;
    case WPC_WATCHDOG:
      break; /* Just ignore for now */
    case WPC_FIRQSRC:
      cpu_set_irq_line(WPC_CPUNO, M6809_FIRQ_LINE, CLEAR_LINE);
      break; /* FIRQ Ack */
    case WPC_IRQACK:
      cpu_set_irq_line(WPC_CPUNO, M6809_IRQ_LINE, CLEAR_LINE);
      break;
    case DMD_PAGE3800: /* set the page that is visible at 0x3800 */
      cpu_setbank(2, memory_region(WPC_MEMREG_DMD) + (data & 0x0f) * 0x200); break;
    case DMD_PAGE3A00: /* set the page that is visible at 0x3A00 */
      cpu_setbank(3, memory_region(WPC_MEMREG_DMD) + (data & 0x0f) * 0x200); break;
    case DMD_FIRQLINE: /* set the line to generate FIRQ at. */
      break;
    case DMD_VISIBLEPAGE: /* set the visible page */
      break;
    case WPC_RTCHOUR:
      locals.ltime = (locals.ltime % 60) + data*60;
      break;
    case WPC_RTCMIN:
      locals.ltime = (locals.ltime / 60) * 60 + data;
      break;
    case WPC_RTCLOAD:
      locals.time = locals.ltime;
      break;
    case WPC_LED:
      break; /* ignore LED */
    case WPC_PRINTDATA:
      break;
    case WPC_PRINTDATAX:
      if (data == 0) {
        if (wpc_printfile == NULL) {
          char filename[13];

          sprintf(filename,"%s.prt", Machine->gamedrv->name);
          wpc_printfile = osd_fopen(Machine->gamedrv->name,filename,OSD_FILETYPE_MEMCARD,1);
        }
        osd_fwrite(wpc_printfile, &wpc_data[WPC_PRINTDATA], 1);
      }
      break;
    default:
      /*logerror("wpc_w %4x %2x\n", offset+WPC_BASE, data);*/
      break;
  }
  wpc_data[offset] = data;
}

/*--------------------------
/ Security chip simulation
/---------------------------*/
static int wpc_pic_r(void) {
  int ret = 0;
  if (locals.pic.lastW == 0x0d)
    ret = locals.pic.count;
  else if ((locals.pic.lastW >= 0x16) && (locals.pic.lastW <= 0x1f))
    ret = wpc_globals.swMatrix[locals.pic.lastW-0x15];
  else if ((locals.pic.lastW & 0xf0) == 0x70) {
    ret = locals.pic.sData[locals.pic.lastW & 0x0f];
    /* update serial number scrambler */
    locals.pic.sNoS = ((locals.pic.sNoS>>4) | (locals.pic.lastW <<4)) & 0xff;
    locals.pic.sData[5]  = (locals.pic.sData[5]  ^ locals.pic.sNoS) + locals.pic.sData[13];
    locals.pic.sData[13] = (locals.pic.sData[13] + locals.pic.sNoS) ^ locals.pic.sData[5];
  }
  return ret;
}

static void wpc_pic_w(int data) {
  if (locals.pic.codeW > 0) {
    if (locals.pic.codeNo[3 - locals.pic.codeW] != data)
      logerror("Wrong code %2x (%2x) sent to pic.",
               data, locals.pic.codeNo[3 - locals.pic.codeW]);
    locals.pic.codeW -= 1;
  }
  else if (data == 0) {
    locals.pic.sNoS = 0xa5;
    locals.pic.sData[5]  = locals.pic.sData[0] ^ locals.pic.sData[15];
    locals.pic.sData[13] = locals.pic.sData[2] ^ locals.pic.sData[12];
    locals.pic.count = 0x20;
  }
  else if (data == 0x20) {
    locals.pic.codeW = 3;
  }
  else if (data == 0x0d) {
    locals.pic.count = (locals.pic.count - 1) & 0x1f;
  }

  locals.pic.lastW = data;
}

/*-------------------------
/  Generate IRQ interrupt
/--------------------------*/
static int wpc_irq(void) {
  /*-- update rtc clock --*/
  if (++locals.ticks >= 60*WPC_IRQFREQ)
    { locals.time += 1; locals.ticks = 0; }
  cpu_set_irq_line(WPC_CPUNO, M6809_IRQ_LINE, HOLD_LINE);
  return 0;
}

static void wpc_init(void) {
                              /*128K  256K        512K        768K       1024K*/
  static int romLengthMask[] = {0x07, 0x0f, 0x00, 0x1f, 0x00, 0x00, 0x00, 0x3f};
  int romLength = memory_region_length(WPC_MEMREG_ROM);

  /*-- clear static variables --*/
  memset(&locals, 0, sizeof(locals));
  memset(&wpc_globals, 0, sizeof(wpc_globals));

  /*-- Initialize Sound Command Variables --*/
  snd_cmd_init();
  
  /*-- there's no game exit function so we do this on startup of the next game --*/
  if (wpc_printfile) { osd_fclose(wpc_printfile); wpc_printfile = NULL; }

  locals.pageMask = romLengthMask[((romLength>>17)-1)&0x07];

  /* the non-paged ROM is at the end of the image. move it into the CPU region */
  memcpy(memory_region(WPC_MEMREG_CPU) + 0x8000,
         memory_region(WPC_MEMREG_ROM) + romLength - 0x8000, 0x8000);

  /*-- sync counter with vblank --*/
  locals.vblankCount = 1;

  if (wpc_gameData == NULL)
    { logerror("wpc_init: gameData not initialised\n"); return; }

  /*-- init sound (if present) --*/
  if (USESOUND) {
    WPC_gWPC95 = (wpc_gameData->wpcGen == WPC_GEN95);
    if (wpc_gameData->wpcGen < WPC_GENDCS) wpcs_init();
    else                                   dcs_init();
  }

  /*-- init switch matrix --*/
  memcpy(&locals.invSw, wpc_gameData->invSw, sizeof(wpc_gameData->invSw));
  memcpy(wpc_globals.swMatrix, locals.invSw, sizeof(wpc_globals.swMatrix));
  wpc_globals.swMatrix[2] |= 0x08; /* Always closed switch */

  /*-- init security chip (if present) --*/
  if ((wpc_gameData->wpcGen >= WPC_GENSECURITY) &&
      (wpc_gameData->serialNo))
    wpc_serialCnv(wpc_gameData->serialNo, locals.pic.sData,
                  locals.pic.codeNo);

  if (wpc_gameData->simData) {
    /*-- init the simulator --*/
    int inports[WPC_MAXPORTS];
    int ii;

    for (ii = 0; ii < WPC_CUSTINPORT; ii++)
      inports[ii] = readinputport(ii);

    locals.simAvail = wpcsim_init((wpcsim_tSimData *)wpc_gameData->simData,
                                  inports);
  }
  /*-- command line options --*/
  locals.DMDsize = (options.antialias ||
                    wpc_gameData->wpcGen == WPC_GENALPHA) ? 2 : 1;
  locals.dmdOnly = !options.translucency;
  if (locals.dmdOnly) {
    if (locals.DMDsize > 1)  set_visible_area(0,255,0,65);
    else                     set_visible_area(0,127,1,32);
  }
  if (options.cheat) {
    /*-- speed up startup by disable checksum --*/
    *(memory_region(WPC_MEMREG_CPU) + 0xffec) = 0x00;
    *(memory_region(WPC_MEMREG_CPU) + 0xffed) = 0xff;
  }
}

/*-----------------------------------------------
/ Load/Save static ram
/ Most of the SRAM is cleared on startup but the
/ non-cleared part differs from game to game.
/ Better save it all
/ The RAM size was changed from 8K to 16K starting with
/ DCS generation
/-------------------------------------------------*/
void wpc_nvram(void *file, int write) {
  int length = (wpc_gameData->wpcGen >= WPC_GENDCS) ? 0x3800 : 0x2000;
  if (write)  /* save nvram */
    osd_fwrite(file, memory_region(WPC_MEMREG_CPU), length);
  else if (file) /* load nvram */
    osd_fread(file, memory_region(WPC_MEMREG_CPU), length);
  else        /* first time */
    memset(memory_region(WPC_MEMREG_CPU), 0xff, length);
}

/*--------------------------
/ Write text on the screen
/---------------------------*/
void wpc_textOut(char *buf, int length, int x, int y, int color) {
  int ii, l;

  l = strlen(buf);
  for (ii = 0; ii < length; ii++) {
    char c = (ii >= l) ? ' ' : buf[ii];

    drawgfx(Machine->scrbitmap, Machine->uifont, c, color, 0, 0,
            x + ii * Machine->uifont->width, y+WPC_FIRSTSIMROW, 0,
            TRANSPARENCY_NONE, 0);
  }
}

/*-----------------------------------
/ Write formatted text on the screen
/------------------------------------*/
void CLIB_DECL wpc_textOutf(int x, int y, int color, char *text, ...) {
  va_list arg;
  char buf[100];
  char *bufPtr = buf;

  va_start(arg, text);
  vsprintf(buf, text, arg);
  va_end(arg);

  while (*bufPtr) {
    drawgfx(Machine->scrbitmap, Machine->uifont, *bufPtr++, color, 0, 0,
            x, y+WPC_FIRSTSIMROW, 0,
            TRANSPARENCY_NONE, 0);
    x += Machine->uifont->width;
  }
}

/*---------------------------------------
/  Read the current switch value
/
/  This function returns TRUE for active
/  switches even if the switch is active low.
/-----------------------------------------*/
int wpc_getSw(int swNo) {
  return (wpc_globals.swMatrix[swNo/10] ^ locals.invSw[swNo/10]) & (1<<(swNo%10-1));
}

/*----------------------
/  Set/reset a switch
/-----------------------*/
void wpc_setSw(int swNo, int value) {
  wpc_globals.swMatrix[swNo/10] &= ~(1<<(swNo%10-1)); /* clear the bit first */
  wpc_globals.swMatrix[swNo/10] |=  ((value ? 0xff : 0) ^ locals.invSw[swNo/10]) & (1<<(swNo%10-1));
}

/*-------------------------
/  update active low/high
/-------------------------*/
void wpc_updInvSw(int swNo, int inv) {
  int bit = (1 << (swNo%10-1));

  if (inv)
    inv = bit;
  if ((locals.invSw[swNo/10] ^ inv) & bit) {
    locals.invSw[swNo/10] ^= bit;
    wpc_globals.swMatrix[swNo/10] ^= bit;
  }
}

/*-------------------------------------
/  Read the status of a solenoid
/  For the standard solenoids this is
/  the "smoothed" value
/--------------------------------------*/
int wpc_getSol(int solNo) {
  if (solNo <= WPC_STDSOL)
    return (locals.solData & (1<<(solNo-1))) > 0;
  else if (solNo < WPC_FLIPPERSOL)
    return wpcsim_getSol(solNo); /* simulated */
  else if (solNo < WPC_CUSTSOL) {
    int mask = 1<<(solNo - WPC_FLIPPERSOL);
    /*-- flipper coils --*/
    if      ((solNo == sLRFlip) && (wpc_gameData->flippers & FLIP_SOL(FLIP_LR)))
      mask = 0x03;
    else if ((solNo == sLLFlip) && (wpc_gameData->flippers & FLIP_SOL(FLIP_LL)))
      mask = 0x0c;
    else if ((solNo == sURFlip) && (wpc_gameData->flippers & FLIP_SOL(FLIP_UR)))
      mask = 0x30;
    else if ((solNo == sULFlip) && (wpc_gameData->flippers & FLIP_SOL(FLIP_UL)))
      mask = 0xc0;
    return locals.solFlip & mask;
  }
  else if (wpc_gameData->getSol)
    return wpc_gameData->getSol(solNo);
  return 0;
}

/*---------------------------------------------------------------------------*/
static void wpc_updateSw(void) {
  int inports[WPC_MAXPORTS];
  int ii;

  /*--------------------
  / read standard keys
  /--------------------*/
  for (ii = 0; ii < WPC_CUSTINPORT; ii++)
    inports[ii] = readinputport(ii);

  /*-- update coindoor switches --*/
  wpc_globals.swMatrix[WPC_COINDOORSWCOL] = inports[WPC_COMINPORT] & 0xff;

  /*-- handle flippers--*/
  { /*-- buttons **/
    int flip = wpc_gameData->flippers;
    if (flip & FLIP_SW(FLIP_LL))     /* have LL switch */
      wpc_setSw(FLIP_SWL(flip) ?
                FLIP_SWL(flip) : swLLFlip, inports[WPC_COMINPORT] & WPC_COMLLFLIPKEY);
    if (flip & FLIP_SW(FLIP_UL))     /* have UL switch */
      wpc_setSw(swULFlip,
                inports[WPC_COMINPORT] & ((flip & FLIP_BUT(FLIP_UL)) ? WPC_COMULFLIPKEY :
                                                                       WPC_COMLLFLIPKEY));
    if (flip & FLIP_SW(FLIP_LR))     /* have LR switch */
      wpc_setSw(FLIP_SWR(flip) ?
                FLIP_SWR(flip) : swLRFlip, inports[WPC_COMINPORT] & WPC_COMLRFLIPKEY);
    if (flip & FLIP_SW(FLIP_UR))     /* have UR swicth */
      wpc_setSw(swURFlip,
                inports[WPC_COMINPORT] & ((flip & FLIP_BUT(FLIP_UR)) ? WPC_COMURFLIPKEY :
                                                                       WPC_COMLRFLIPKEY));
    /*-- EOS switches --*/
    if (wpc_gameData->flippers & FLIP_SOL(FLIP_UL)) {
      if (wpc_getSol(sULFlip)) locals.flipTimer[0] += 1;
      else                     locals.flipTimer[0] = 0;
      wpc_setSw(swULFlipEOS, (locals.flipTimer[0] >= WPC_FLIPSTROKETIME));
    }
    if (wpc_gameData->flippers & FLIP_SOL(FLIP_UR)) {
      if (wpc_getSol(sURFlip)) locals.flipTimer[1] += 1;
      else                     locals.flipTimer[1] = 0;
      wpc_setSw(swURFlipEOS, (locals.flipTimer[1] >= WPC_FLIPSTROKETIME));
    }
    if (wpc_gameData->flippers & FLIP_SOL(FLIP_LL)) {
      if (wpc_getSol(sLLFlip)) locals.flipTimer[2] += 1;
      else                     locals.flipTimer[2] = 0;
      wpc_setSw(swLLFlipEOS, (locals.flipTimer[2] >= WPC_FLIPSTROKETIME));
    }
    if (wpc_gameData->flippers & FLIP_SOL(FLIP_LR)) {
      if (wpc_getSol(sLRFlip)) locals.flipTimer[3] += 1;
      else                     locals.flipTimer[3] = 0;
      wpc_setSw(swLRFlipEOS, (locals.flipTimer[3] >= WPC_FLIPSTROKETIME));
    }
  }

  /*-- check if we should use simulator keys --*/
  if (!locals.simAvail || inports[WPC_SIMINPORT] & WPCSIM_SWITCHKEY) {
    /*-- simulator keys disabled, use row+column keys --*/
    static int lastRow = 0, lastCol = 0;
    int row = 0, col = 0;

    if (((inports[WPC_MANSWINPORT] & WPC_MANSWCOLUMNS) == 0) ||
        ((inports[WPC_MANSWINPORT] & WPC_MANSWROWS) == 0))
      lastRow = lastCol = 0;
    else {
      int bit = 0x0101;

      for (ii = 0; ii < 8; ii++) {
        if (inports[WPC_MANSWINPORT] & WPC_MANSWCOLUMNS & bit) col = ii+1;
        if (inports[WPC_MANSWINPORT] & WPC_MANSWROWS    & bit) row = ii+1;
        bit <<= 1;
      }
      if ((col != lastCol) || (row != lastRow)) {
        wpc_setSw(col*10+row, !wpc_getSw(col*10+row));
        lastCol = col; lastRow = row;
      }
    }
  }
  /*-- check standard keys --*/
  if (wpc_gameData->comSw.start)
    wpc_setSw(wpc_gameData->comSw.start,    inports[WPC_COMINPORT] & WPC_COMSTARTKEY);
  if (wpc_gameData->comSw.tilt)
    wpc_setSw(wpc_gameData->comSw.tilt,     inports[WPC_COMINPORT] & WPC_COMTILTKEY);
  if (wpc_gameData->comSw.sTilt)
    wpc_setSw(wpc_gameData->comSw.sTilt,    inports[WPC_COMINPORT] & WPC_COMSTILTKEY);
  if (wpc_gameData->comSw.coinDoor)
    wpc_setSw(wpc_gameData->comSw.coinDoor, inports[WPC_COMINPORT] & WPC_COMCOINDOORKEY);
  if (wpc_gameData->comSw.shooter)
    wpc_setSw(wpc_gameData->comSw.shooter,  inports[WPC_SIMINPORT] & WPCSIM_SHOOTERKEY);

  if (locals.simAvail)
    wpcsim_run(inports, (inports[WPC_SIMINPORT] & WPCSIM_SWITCHKEY) == 0,
               (WPCSIM_BALLS(inports[WPC_SIMINPORT])));

#ifdef MAME_DEBUG
  /* Press W and E at the same time to insert a mark in logfile */
  if ((inports[WPC_MANSWINPORT] & 0x06) == 0x06)
    logerror("\nLogfile Mark\n");
#endif /* MAME_DEBUG */
}

static void wpc_serialCnv(char no[21], unsigned char pic[16],
                          unsigned char code[3]) {
  int x;

  pic[10] = 0x12; /* whatever */
  pic[2]  = 0x34; /* whatever */
  x = (no[5] - '0') + 10*(no[8]-'0') + 100*(no[1]-'0');
  x += pic[10]*5;
  x *= 0x001bcd; /*   7117 = 11*647         */
  x += 0x01f3f0; /* 127984 = 2*2*2*2*19*421 */
  pic[1]  = x >> 16;
  pic[11] = x >> 8;
  pic[9]  = x;
  x = (no[7]-'0') + 10*(no[9]-'0') + 100*(no[0]-'0') + 1000*(no[18]-'0') + 10000*(no[2]-'0');
  x += 2*pic[10] + pic[2];
  x *= 0x0000107f; /*    4223 = 41*103     */
  x += 0x0071e259; /* 7463513 = 53*53*2657 */
  pic[7]  = x >> 24;
  pic[12] = x >> 16;
  pic[0]  = x >> 8;
  pic[8]  = x;
  x = (no[17]-'0') + 10*(no[6]-'0') + 100*(no[4]-'0') + 1000*(no[19]-'0');
  x += pic[2];
  x *= 0x000245; /*   581 = 7*83         */
  x += 0x003d74; /* 15732 =2*2*3*3*19*23 */
  pic[3]  = x >> 16;
  pic[14] = x >> 8;
  pic[6]  = x;
  x = ('9'-no[11]) + 10*('9'-no[12]) + 100*('9'-no[13]) + 1000*('9'-no[14]) + 10000*('9'-no[15]);
  pic[15] = x >> 8;
  pic[4]  = x;

  x = 100*(no[0]-'0') + 10*(no[1]-'0') + (no[2]-'0');
  x = (x >> 8) * (0x100*no[17] + no[19]) + (x & 0xff) * (0x100*no[18] + no[17]);
  code[0] = x >> 16;
  code[1] = x >> 8;
  code[2] = x;
}

#ifdef OLDANTIALIAS
static void copyBM(struct osd_bitmap *bitmap, int antialias) {
  UINT8  **dest8  = (UINT8  **)bitmap->line;
  UINT16 **dest16 = (UINT16 **)bitmap->line;
  int ii,jj;

  int how = 8*((bitmap->depth-1)/8) +
            4*((Machine->orientation & ORIENTATION_SWAP_XY) > 0) +
            2*((Machine->orientation & ORIENTATION_FLIP_Y) > 0) +
              ((Machine->orientation & ORIENTATION_FLIP_X) > 0);

  if (antialias) {
    for (ii = 0; ii < 31; ii++) {
      for (jj = 0; jj < 127; jj++) {
        wpc_bm->line[2*ii]  [2*jj+1] = (wpc_bm->line[2*ii][2*jj]   + wpc_bm->line[2*ii]  [2*jj+2] - 2) + 17;
        wpc_bm->line[2*ii+1][2*jj]   = (wpc_bm->line[2*ii][2*jj]   + wpc_bm->line[2*ii+2][2*jj]   - 2) + 17;
        wpc_bm->line[2*ii+1][2*jj+1] = (wpc_bm->line[2*ii][2*jj]   + wpc_bm->line[2*ii+2][2*jj] +
                                        wpc_bm->line[2*ii][2*jj+2] + wpc_bm->line[2*ii+2][2*jj+2] - 4)*2/5 + 17;
      }
    }
  }

  for (ii = 0; ii < WPC_SCREENX; ii++) {
    for (jj = 0; jj < WPC_SCREENY; jj++) {
      int data = Machine->pens[wpc_bm->line[ii][jj]];
      if (data) {
        switch (how) {
          case 0x0: /*  8 bit */
            dest8[ii][jj] = data; break;
          case 0x8: /* 16 bit */
            dest16[ii][jj] = data; break;
          case 0x1: /*  8 bit flipX */
            dest8[ii][WPC_SCREENY-1-jj] = data; break;
          case 0x9: /* 16 bit flipX */
            dest16[ii][WPC_SCREENY-1-jj] = data; break;
          case 0x2: /*  8 bit flipY */
            dest8[WPC_SCREENX-1-ii][jj] = data; break;
          case 0xa: /* 16 bit flipY */
            dest16[WPC_SCREENX-1-ii][jj] = data; break;
          case 0x3: /*  8 bit flipX flipY */
            dest8[WPC_SCREENX-1-ii][WPC_SCREENY-1-jj] = data; break;
          case 0xb: /* 16 bit flipX flipY */
            dest16[WPC_SCREENX-1-ii][WPC_SCREENY-1-jj] = data; break;
          case 0x4: /*  8 bit swapXY */
            dest8[jj][ii] = data; break;
          case 0xc: /* 16 bit swapXY */
            dest16[jj][ii] = data; break;
          case 0x5: /*  8 bit swapXY flipX */
            dest8[jj][WPC_SCREENX-1-ii] = data; break;
          case 0xd: /* 16 bit swapXY flipX */
            dest16[jj][WPC_SCREENX-1-ii] = data; break;
          case 0x6: /*  8 bit swapXY flipY */
            dest8[WPC_SCREENY-1-jj][ii] = data; break;
          case 0xe: /* 16 bit swapXY flipY */
            dest16[WPC_SCREENY-1-jj][ii] = data; break;
          case 0x7: /*  8 bit swapXY flipX flipY */
            dest8[WPC_SCREENY-1-jj][WPC_SCREENX-1-ii] = data; break;
          case 0xf: /* 16 bit swapXY flipX flipY */
            dest16[WPC_SCREENY-1-jj][WPC_SCREENX-1-ii] = data; break;
        } /* switch */
      }
    }
  }
}
#endif /* ANTIALIAS */
